/* VEMS ROUND Copyright 2005 Marcell Gal (XDSL Ltd Hungary) */

/*!
 * \addtogroup g_serial_comm Serial Communication Routines
 */
/*@{*/


#include "comm.h"
#include "adc.h"
#include "wbo2.h"
#include "ringbuf.h"
#include "arithm.h"
#include "tlc5920.h"
#include "storage.h"
#include "egt.h"
#include "round.h"
#include "firmware_revision.h"
#include "aim.h"

#include <avr/io.h>
#include <inttypes.h>
/* #include <avr/signal.h> */
#include <avr/interrupt.h>
#include <avr/sfr_defs.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

struct cmd_queue_t cmd_queue[CMD_QUEUE_SIZE];
uint8_t cmd_queue_tail, cmd_queue_head;
uint8_t comm_state;
uint8_t comm_prevrx;
uint8_t rx_offset;
uint8_t rx_cmd;
uint8_t comm_TX_plainbytes[comm_TX_plain_SIZE];

/* fill these somewhere later */
uint16_t miscADC=0; /* this would be the buttonADC input, with button-filtered out ? */
uint8_t atomic_otherbyte; /* snapshot of the other byte when one byte of a 16 bit variable is sent */
/* comm.c adapted from v3 stable comm.c */

uint8_t aim_state;

void comm_cmd(uint8_t rx_data)
{
	switch (rx_data) {
		case 'A': /* send all realtime values */
			pushfunc(&sendRTVar, MT_RTVAR_COUNT);
		break;
		case 'B': /* store to eeprom */
			save_variables();
		break;
		case 'S': /* send textual signature */
			pushfuncpar(&srl_generic_P, sizeof(ROUND_VERSION_STR)-1, (uint8_t*)PSTR(ROUND_VERSION_STR));
		break;

		case 'c': /* config read */
			pushfunc(&sendConfigVar, sizeof(struct config_t));
		break;

		case 'C': /* config write */
			comm_state = OFFSET;
			rx_cmd = 'C';
		break;

		default:
		break;
	}
}

void comm(uint8_t rx_data)
{
	switch (comm_state) {
		case OFFSET:
			rx_offset = rx_data;  /* the offset for storing the new byte */
			comm_state = VALUE;
			break;

		case VALUE:
			switch (rx_cmd) {

				case 'C':
					storeConfigPar(rx_offset, rx_data);
					break;

				default:
				break;
			}
			comm_state = CMD; /* back to normal mode */
			break;

        case AIM_DATA:
            aim_state = aim_feed(rx_data);
            if(aim_state == AIM_OK || aim_state == AIM_ERROR){ /* what else can be ? */
                comm_state = CMD; /* back to normal mode */
            }
            break;
	
	case AIM_CANDIDATE:
		if (rx_data == AIM_CONST){
			aim_feed(comm_prevrx);
                	aim_feed(AIM_CONST);
               		comm_state = AIM_DATA; /* go AIM_DATA mode */
		} else { /* it's not AIM */
			comm_state = CMD; /* command mode for prev byte */
			comm_cmd(comm_prevrx);
			/* prevent infinite recursion - should never be true anyway:
			if( comm_state == AIM_CANDIDATE ){
				comm_state = CMD;
			} */
			comm(rx_data);
		}
		break;
	default:
	/* AIM design is stupid, because the flagword is not at the beginning.
		channel, 0xA3 flagword, data, checksum
		This makes it hell when other packets also travel on same line
	*/
		/*if( (config.segment0 == 0xA5 || config.segment3 == 0xA5) ){*/
		if( (config.units & 0x80 ) ){
			comm_prevrx = rx_data;
			/* can AIM channel be A3 ? No, max channel is 125 */
			comm_state = AIM_CANDIDATE;
                	return;
		} else { /* command mode */
			comm_cmd(rx_data);
		}
	}
}

/***************************************************************************/
/*** sendConfigVar                                                       ***/
/***************************************************************************/
uint8_t sendConfigVar(uint8_t i) {
	if (i < sizeof(struct config_t)) {
		return *( (uint8_t *)&config + i );
	} else {
		return 0;
	}
}


/***************************************************************************/
/*** storeConfigPar                                                       ***/
/***************************************************************************/
void storeConfigPar(uint8_t offset, uint8_t value) {
	if (offset < sizeof(struct config_t)) {
		*((uint8_t*)&config + offset) = value;
	}
}


/* ========================================
* Note: little endian protocol would make it possible to send DIRECTLY FROM SRAM,
* and save a lot of flash. Any future big endian processor (if ever applied)
* would be powerful enough to send without problem
*/
uint8_t sendRTVar(uint8_t i) {
/*	uint8_t a, b; */

	switch (i) {

		case 0:
			return (wbo2.current_lambda >> 8);

		case 1:
			return (uint8_t) wbo2.current_lambda;

		case 2:
			return (wbo2.current_o2 >> 8);

		case 3:
			return (uint8_t) wbo2.current_o2;

		case 4:
			return (runtime.current_egt >> 8);

		case 5:
			return (uint8_t) runtime.current_egt;

		case 6:
			return runtime.rpm_vhr >> 8;

		case 7:
			return (uint8_t) runtime.rpm_vhr;

		case 8:
			return miscADC >> 8;

		case 9:
			return (uint8_t) miscADC;

		case 10:
			return vout_pwm.designated_ratio;

		case 11:
			return boost_pwm.designated_ratio;

		case 12: /* what is this ? */
			return runtime.status;

		case 13:
			return runtime.internal_temp;

		case 30:
			return runtime.analin >> 8;
		case 31:
			return (uint8_t)runtime.analin;

		default:
		/* 14 .. 29 are 8 (filtered 16bit) ADC readings. See adc.h
		 * Eg. BUTTON_ADC = 4 so 22=14+4*2 is high byte, 23=14+4*2+1 is low byte
		 */
			if( (i & 1) == 0 ){
				uint16_t sample = sensor_filt[ (i-14) >> 1 ];
				/* (((void*)sensor_filt) + i) would save flash */
				atomic_otherbyte = (uint8_t)sample;
				return sample >> 8;
			} else {
				return atomic_otherbyte;
			}
	}
}


/************************************
* generic serial logging function.
* uses private from the actual command as a pointer (array base).
* Unfortunately we need to know the string-length in advance
* For constant(-length) strings this is OK.
* We could set a (global or static variable is simplest) flag
* to notify the caller about the end of transmission.
* This way other comm functions could be left untouched.
************************************/
uint8_t srl_generic(uint8_t addr)
{
	return (cmd_queue[cmd_queue_head].p1.private)[addr];
}

uint8_t srl_byte(uint8_t addr)
{
  uint8_t b=(cmd_queue[cmd_queue_head].p1.private)[0];
  if (addr)
    b=b>>4;
  else
    b&=0x0F;
  if (b<10)
    return '0'+b;
  return 'A'+b-10;
}

/* same as above, but private points to PROGMEM (flash), eg. PSTR("...") */
uint8_t srl_generic_P(uint8_t addr)
{
	return pgm_read_byte(&(cmd_queue[cmd_queue_head].p1.private[addr]));
}

/***************************************************************************/
/*** pushfunc                                                            ***/
/***************************************************************************/
uint8_t pushfunc(uint8_t (*f)(uint8_t i), uint8_t len) {

	uint8_t next_tail;

	next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
	if (next_tail == cmd_queue_head) {
		return -1; /* whoops, queue is full */
	}

	cmd_queue[cmd_queue_tail].func = f;
	cmd_queue[cmd_queue_tail].len = len;
	cmd_queue[cmd_queue_tail].count = 0;
	cmd_queue[cmd_queue_tail].p1.private = 0;

	cmd_queue_tail = next_tail;

	return 0;
}


/**********************************************
* same as above, but parameter can be specified
*/
uint8_t pushfuncpar2(uint8_t (*f)(uint8_t i), uint8_t len, const uint8_t * par, uint8_t par2) {

	uint8_t next_tail;

	next_tail = (cmd_queue_tail + 1) % CMD_QUEUE_SIZE;
	if (next_tail == cmd_queue_head) {
		return -1; /* whoops, queue is full */
	}

	cmd_queue[cmd_queue_tail].func = f;
	cmd_queue[cmd_queue_tail].len = len;
	cmd_queue[cmd_queue_tail].count = 0;
	cmd_queue[cmd_queue_tail].p1.private = (void*)par;
	cmd_queue[cmd_queue_tail].p2.privch.a = par2;
	cmd_queue_tail = next_tail;

	return 0;
}

/* compatibility function with less parameters */
uint8_t pushfuncpar(uint8_t (*f)(uint8_t i), uint8_t len, const uint8_t * par) {

	return pushfuncpar2(f, len, par, 0);
}


void enable_transmit_interrupt(void) {
	/* enable transmit interrupt */
	UCSR0B |= _BV(UDRIE0);
}


/***************************************************************************/
/*** initUART                                                            ***/
/***************************************************************************/
void initUART(void) {
	/* reset the baudrate flag. 0-d by gcc at startup, but ... */
	comm_state = CMD;

	/*
	 * The whole init() happens in cli()
	 * stay in clear and disable interrupts for initialization with interrupt driven communications
	 * as per atmega168 datasheet page 174
	 * cli();
	 */

	/* rx enable, tx enable */
	UCSR0B = _BV(RXCIE0) | _BV(RXEN0) | _BV(TXEN0);
	/* 8 databit, 1 stopbit */
	UCSR0C = _BV(UCSZ01) | _BV(UCSZ00);

	/* according to atmega88/atmega168 datasheet page#185, for baudrate of 19.2k/s at 16MHz
	   UBRR0 = 51, error: 0.2%
	*/
	UBRR0H = 0; /* not an Octal value any more */
	UBRR0L = 51;

	/* DON'T enable interrupts, here ! This breaks other initializations.
	We stay in cli() until everything is initialized
	sei();
	*/

	cmd_queue_tail = cmd_queue_head = 0;
}


/***************************************************************************/
/*** INTERRUPT: UART rx complete ***/
/***************************************************************************/
SIGNAL(SIG_USART_RECV) {

	/* check status register for frame or overrun errors */
	if ( UCSR0A & (_BV(FE0) | _BV(DOR0)) ) {
		uint8_t dummy;
		dummy = UDR0; /* just read uart and throw away the result */
		return;
	} else {
		ringbuf_push(&comm_recbuf, UDR0);
	}
}

uint8_t srl_direct( uint8_t i ){

	switch(i){
		/* case 0: return 10; */
		/* case 1: return 'd'; */
		case 0: {
			uint16_t t = (uint16_t) cmd_queue[cmd_queue_tail].p1.private;
			return t>>8;
		}
		case 1: {
			uint16_t t = (uint16_t) cmd_queue[cmd_queue_tail].p1.private;
			return t & 0xff;
		}
		default: return 'z';
	}
}

/***************************************************************************/
/*** INTERRUPT: UART tx empty ***/
/***************************************************************************/
SIGNAL(SIG_USART_DATA) {
	uint16_t d;

	d = ringbuf_shift(&comm_sndbuf);

	if (d>>8) { /* buffer empty */
		UCSR0B &= ~_BV(UDRIE0);    /* disable TX interrupts */
	} else {
		UDR0 = d;
	}

}

/*@}*/


